"""
model evaluation
"""
from .data_prep import load_eval_set_as_df, split_eval_set
from .log_metrics import log_metrics
from .serving import get_base_model_details

__all__ = ['log_metrics', 'load_eval_set_as_df', 'split_eval_set', 'get_base_model_details']



"""Log eval metrics as FT RunEvent"""

from typing import List

from databricks.model_training.api.engine import get_return_response, run_singular_mapi_request
from databricks.model_training.types import EvalMetricDict, TrainingRun

QUERY_FUNCTION = 'createRunEvent'
VARIABLE_DATA_NAME = 'createRunEventData'

QUERY = f"""
mutation CreateRunEvent(${VARIABLE_DATA_NAME}: CreateRunEventInput!) {{
    {QUERY_FUNCTION}({VARIABLE_DATA_NAME}: ${VARIABLE_DATA_NAME}) {{
        eventData
        eventType
        executionId
        id
        runId
    }}
}}"""


def log_metrics(
    run_name: str,
    dataset_id: str,
    eval_metrics: List[EvalMetricDict],
) -> None:
    """
    Logs evaluation metrics as RunEvent

    Args:
        run_name (str): The name of the run to log the metrics for.
        dataset_id (str): Some identifier for the dataset. This can be a name, filepath, etc.
        eval_metrics (List[EvalMetric]): A list of evaluation metrics to log.
    """
    if len(eval_metrics) == 0:
        raise ValueError('No evaluation metrics provided.')
    for metric in eval_metrics:
        if not all(key in metric for key in ['metric_name', 'score', 'greater_is_better']):
            raise ValueError(
                "Each evaluation metric must contain 'metric_name', 'score', and 'greater_is_better' fields.")

    metrics = [{
        'metricName': metric['metric_name'],
        'score': metric['score'],
        'greaterIsBetter': metric['greater_is_better'],
        'baselineScore': metric.get('baseline_score'),
    } for metric in eval_metrics]
    variables = {
        VARIABLE_DATA_NAME: {
            'eventData': {
                'evalDatasetId': dataset_id,
                'metrics': metrics,
            },
            'eventType': 'LOG_EVALUATION_METRIC',
            'runName': run_name,
        }
    }

    response = run_singular_mapi_request(
        query=QUERY,
        query_function=QUERY_FUNCTION,
        return_model_type=TrainingRun,
        variables=variables,
    )

    try:
        run = get_return_response(response)
    except Exception as e:
        raise ValueError(f'Failed to log evaluation results. {e}') from e

    print(f'Logged evaluation results for run {run.name}.')



"""Cancel model training runs"""

import logging
from typing import Any, Dict, List, Union

from databricks.model_training.api.engine import get_return_response, run_plural_mapi_request
from databricks.model_training.api.exceptions import DatabricksModelTrainingRequestError
from databricks.model_training.types.training_run import TrainingRun

logger = logging.getLogger(__name__)

QUERY_FUNCTION = 'stopFinetunes'
VARIABLE_DATA_NAME = 'getFinetunesData'
OPTIONAL_DATA_NAME = 'stopFinetunesData'
QUERY = f"""
mutation StopFinetunes(${VARIABLE_DATA_NAME}: GetFinetunesInput!, ${OPTIONAL_DATA_NAME}: StopFinetunesInput) {{
  {QUERY_FUNCTION}({VARIABLE_DATA_NAME}: ${VARIABLE_DATA_NAME}, {OPTIONAL_DATA_NAME}: ${OPTIONAL_DATA_NAME}) {{
    id
    name
    status
    createdById
    createdByEmail
    createdAt
    updatedAt
    startedAt
    completedAt
    reason
    isDeleted
    details {{
            model
            trainDataPath
            saveFolder
            experimentTracker
        }}
  }}
}}"""


def cancel(experiment_ids: Union[str, List[str]]) -> int:
    """
    Cancel all sweep run(s) by experiment ID(s)

    Args:
        experiment_ids (str): List of experiment IDs to cancel

    Raises:
        DatabricksModelTrainingRequestError: Raised if stopping any of the requested runs failed

    Returns:
        int: The number of training runs canceled
    """
    if not experiment_ids:
        raise ValueError('Must provide valid experiment ID(s) to cancel.')
    if isinstance(experiment_ids, str):
        experiment_ids = [experiment_ids]
    filters = {'mlflowExperimentId': {'in': experiment_ids}}
    variables: Dict[str, Dict[str, Any]] = {
        VARIABLE_DATA_NAME: {
            'filters': filters
        },
        OPTIONAL_DATA_NAME: {
            'reason': 'Canceled by user.'
        }
    }

    try:
        response = run_plural_mapi_request(
            query=QUERY,
            query_function=QUERY_FUNCTION,
            return_model_type=TrainingRun,
            variables=variables,
        )
        canceled_runs = get_return_response(response)
        num_canceled = len(canceled_runs)
        if num_canceled == 0:
            print('No active runs found to cancel.')
        return num_canceled
    except Exception as e:
        raise DatabricksModelTrainingRequestError(
            f'Failed to cancel runs for experiment ID(s) {experiment_ids}. Please make sure the runs \
            have not completed or failed and try again.') from e


def cancel_run(run: Union[str, TrainingRun]) -> int:
    """
    Cancel a model training run.

    Args:
        run (Union[str, TrainingRun]): The training run to cancel

    Raises:
        DatabricksModelTrainingRequestError: Raised if stopping any of the requested runs failed

    Returns:
        int: The number of training runs canceled
    """
    if not run:
        raise ValueError('Must provide valid training run to cancel.')
    requested_run_name = run.name if isinstance(run, TrainingRun) else run
    return cancel_runs(runs=[requested_run_name])


def cancel_runs(runs: Union[List[str], List[TrainingRun]]) -> int:
    """
    Cancel a training run, list of training runs, or all runs under a single MLflow experiment ID without
    deleting them. If the run does not exist or if the run has already terminated, an error will be raised.

    Args:
        runs (Optional[Union[List[str], List[TrainingRun]]]): The
            training run(s) to cancel. Can be a single run or a list of runs.

    Raises:
        DatabricksModelTrainingRequestError: Raised if stopping any of the requested runs failed

    Returns:
        int: The number of training runs cancelled
    """
    if len(runs) < 1:
        raise ValueError('Must provide valid training run(s) to cancel.')
    filters = {}
    run_names_to_cancel = [getattr(run, 'name', run) for run in runs if run]
    filters['name'] = {'in': run_names_to_cancel}

    variables: Dict[str, Dict[str, Any]] = {
        VARIABLE_DATA_NAME: {
            'filters': filters
        },
        OPTIONAL_DATA_NAME: {
            'reason': 'Canceled by user.'
        }
    }

    try:
        response = run_plural_mapi_request(
            query=QUERY,
            query_function=QUERY_FUNCTION,
            return_model_type=TrainingRun,
            variables=variables,
        )
        canceled_runs = get_return_response(response)
        num_canceled = len(canceled_runs)
        if num_canceled == 0:
            print('No active runs found to cancel.')
        else:
            canceled_run_names = ', '.join([run.name for run in canceled_runs])
            not_found = ''
            if num_canceled < len(run_names_to_cancel):
                not_found = f' {len(run_names_to_cancel) - num_canceled} run(s) were not found or already terminated.'
            print(f'Canceled {canceled_run_names} successfully.{not_found}')
        return num_canceled
    except Exception as e:
        raise DatabricksModelTrainingRequestError(f'Failed to cancel runs {runs}. Please make sure the run \
                                                  has not completed or failed and try again.') from e



"""
Input to create a sweep.
"""
import logging
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from mcli.utils.utils_config import strip_nones
from mcli.utils.utils_string_functions import snake_case_to_camel_case

from databricks.model_training.api.exceptions import DatabricksGenAIConfigError

logger = logging.getLogger(__name__)


@dataclass
class SweepInput:
    """
    Input to create a Sweep and its training runs.

    Args:
        task_type (str): Task type to finetune on.
        models (List[str]): A list of models. We currently can only take one model.
        train_data_paths (List[str]): A list of training data paths. We currently can only take one train data path.
        model_registry_path (str): The path to the model registry.
        save_folder (str): The folder to save checkpoints.
        eval_data_path (Optional[str], optional): The path to the evaluation data. Defaults to None.
        training_durations (Optional[List[int]], optional): A list of training durations. Defaults to None.
            We generate learning rates and training durations if both aren't provided and submitted model
            supports auto sweep.
        learning_rates (Optional[List[float]], optional): A list of learning rates. Defaults to None.
            We generate learning rates and training durations if both aren't provided and submitted model
            supports auto sweep.
        context_length (Optional[int], optional): The length of the context. Defaults to None.
        experiment_name (Optional[str], optional): The name of the MLflow experiment to create. Defaults to None.
        experiment_directory (Optional[str], optional): The directory of the MLflow experiment to create.
            Defaults to None.
        custom_weights_path (Optional[str], optional): The path to custom weights. Defaults to None.
        data_prep_config (Optional[Dict], optional): The data preparation configuration. Defaults to None.

    Returns:
        None
    """

    task_type: str
    models: List[str]
    train_data_paths: List[str]
    model_registry_path: str
    save_folder: str
    eval_data_path: Optional[str] = None
    training_durations: Optional[List[int]] = None
    learning_rates: Optional[List[float]] = None
    context_length: Optional[int] = None
    experiment_name: Optional[str] = None
    experiment_directory: Optional[str] = None
    custom_weights_path: Optional[str] = None
    data_prep_config: Optional[Dict] = None

    _required_properties = {'models', 'train_data_paths', 'model_registry_path', 'save_folder'}
    _sweep_properties = {
        'task_type',
        'models',
        'train_data_paths',
        'model_registry_path',
        'save_folder',
        'eval_data_path',
        'training_durations',
        'learning_rates',
        'context_length',
        'experiment_name',
        'experiment_directory',
        'custom_weights_path',
        'data_prep_config',
    }

    @classmethod
    def validate_dict(cls, dict_to_use: Dict, show_unused_warning: bool = True) -> Dict:
        """Validate the sweep input from the provided dictionary.
        """
        sweep_input_dict = strip_nones(dict_to_use)
        unused_keys = set(sweep_input_dict) - cls._sweep_properties
        for key in unused_keys:
            del sweep_input_dict[key]
        missing = cls._required_properties - set(dict_to_use)
        if missing:
            raise DatabricksGenAIConfigError(f'Missing required fields: {", ".join(missing)}',)
        if len(unused_keys) > 0 and show_unused_warning:
            raise DatabricksGenAIConfigError(f'Encountered unknown fields in sweep input: {", ".join(unused_keys)}.')
        return sweep_input_dict

    @classmethod
    def from_dict(cls, dict_to_use: Dict[str, Any], show_unused_warning: bool = True):
        """Load the sweep input from the provided dictionary.
        """
        sweep_input_dict = cls.validate_dict(dict_to_use, show_unused_warning)
        return cls(**sweep_input_dict)

    def to_create_api_input(self) -> Dict[str, Dict[str, Any]]:
        """Converts the TrainConfig object to Compute and Finetune that can be used to create a finetune run
        """
        sweep_input = {}
        sweep_input_dict = strip_nones(self.__dict__)
        for key, val in sweep_input_dict.items():
            translated_key = snake_case_to_camel_case(key)
            sweep_input[translated_key] = val

        return {'sweepInput': sweep_input}



"""
Specifies functions that prepare data for evaluation.
Preparation supports any format that works with Mosaic AI Training.
See https://docs.databricks.com/en/large-language-models/foundation-model-training/data-preparation.html
"""
from typing import Optional, Tuple
from uuid import uuid4

import datasets as hf_datasets
import pandas as pd
from databricks.connect import DatabricksSession
from pyspark.sql import DataFrame

from databricks.model_training.api.utils import _get_host_and_token_from_env, get_cluster_id
from databricks.model_training.api.validation import (SupportedDataFormats, format_path, validate_output_to,
                                                      validate_path)
from databricks.model_training.types.train_config import TrainTaskType

AGENT_EVAL_MAX_SAMPLES = 500


def _load_data_as_df(
    data_path: str,
    data_format: SupportedDataFormats,
    use_serverless: bool = False,
    cluster_id: Optional[str] = None,
) -> DataFrame:
    """
    Loads data from the specified path as a Spark DataFrame.

    Args:
        data_path (str): Path to the data to load
        data_format (SupportedDataFormats): Format of the data. Must be a format supported by Mosaic AI Model Training.
        use_serverless (bool): Whether to use serverless compute or not
        cluster_id (Optional[str]): Cluster ID to use (required if not using serverless)

    Returns:
        DataFrame: A PySpark DataFrame representing the loaded data
    """
    if not use_serverless and not cluster_id:
        raise ValueError('cluster_id is required when not using serverless.')
    if use_serverless and cluster_id:
        raise ValueError('cluster_id provided while use_serverless specified. Please provide only one.')

    host, token = _get_host_and_token_from_env()

    try:
        if use_serverless:
            session_id = str(uuid4())
            spark = DatabricksSession.builder.host(host).token(token).header('x-databricks-session-id',
                                                                             session_id).getOrCreate()
        else:
            spark = DatabricksSession.builder.remote(host=host, token=token, cluster_id=cluster_id).getOrCreate()
    except Exception as e:
        raise RuntimeError(f'Failed to connect to Databricks: {str(e)}') from e

    try:
        if data_format == SupportedDataFormats.DELTA_TABLE:
            df = spark.table(data_path)
        elif data_format == SupportedDataFormats.UC_VOLUME:
            df = spark.read.json(data_path)
        elif data_format == SupportedDataFormats.HF_DATASET:
            dataset = hf_datasets.load_dataset(data_path)
            pd_df = pd.DataFrame(dataset)  # type: ignore
            df = spark.createDataFrame(pd_df)
        return df
    except Exception as e:
        spark.stop()  # clean up the Spark session if there's an error
        raise RuntimeError(f"Failed to load '{data_path}': {str(e)}") from e


def _validate_and_load_data_as_df(
    data_path: str,
    data_path_type: str,
    use_serverless: bool = False,
    cluster_id: Optional[str] = None,
) -> DataFrame:
    """
    Validate the data path and load the data as a DataFrame.

    Args:
        data_path (str): Path to the data to load
        data_path_type (str): Type of data path (e.g., "train_data_path", "eval_data_path"). Used for error messages.
        use_serverless (bool): Whether to use serverless compute or not
        cluster_id (Optional[str]): Cluster ID to use (required if not using serverless)

    Returns:
        DataFrame: A PySpark DataFrame representing the loaded data
    """
    # Not supporting CPT evals. No difference between IFT/CHAT validation so we skip prompting task_type as an arg
    data_format = validate_path(data_path, TrainTaskType.INSTRUCTION_FINETUNE, data_path_type)
    df = _load_data_as_df(data_path, data_format, use_serverless, cluster_id)
    return df


def load_eval_set_as_df(eval_data_path: str,) -> DataFrame:
    """
    Loads the evaluation dataset as a Spark DataFrame.

    Args:
        eval_data_path (str): Path to the evaluation dataset. The format should be supported by Mosaic AI Training.

    Returns:
        DataFrame: A PySpark DataFrame representing the loaded evaluation data
    """
    return _validate_and_load_data_as_df(eval_data_path, 'eval_data_path', cluster_id=get_cluster_id())


def split_eval_set(
    train_data_path: str,
    output_to: str,
    eval_split_ratio: float = 0.1,
    max_eval_samples: int = 500,
    seed: Optional[int] = None,
    use_serverless: bool = False,
) -> Tuple[str, str]:
    """
    Splits the dataset into training and evaluation sets. Writes the splits to new Delta tables.

    Args:
        train_data_path (str): Path to the training dataset. The format should be supported by Mosaic AI Training.
        output_to (str): Path to save the training and evaluation splits. This should be in the format 'schema.catalog'.
        eval_split_ratio (float): Ratio of the dataset to use for evaluation. The remainder will be used for training
        max_eval_samples (int): Maximum number of samples to include in the eval set.
        seed (int): Random seed for splitting the dataset
    """
    validate_output_to(output_to)
    if not 0 < eval_split_ratio < 1:
        raise ValueError('eval_split_ratio must be between 0 and 1.')
    if max_eval_samples <= 0:
        raise ValueError(f'max_eval_samples must be a positive integer, got {max_eval_samples}.')
    if max_eval_samples > AGENT_EVAL_MAX_SAMPLES:
        print(
            'max_eval_samples exceeds the maximum allowed value of ' \
            f'{AGENT_EVAL_MAX_SAMPLES}. Setting to {AGENT_EVAL_MAX_SAMPLES}.'
        )
        max_eval_samples = AGENT_EVAL_MAX_SAMPLES

    train_data_path = format_path(train_data_path)

    cluster_id = get_cluster_id() if not use_serverless else None

    # Databricks Default Storage requires serverless compute
    spark_df = _validate_and_load_data_as_df(train_data_path, 'train_data_path', use_serverless, cluster_id)

    total_rows = spark_df.count()
    batch_size = 1 << 30
    num_partitions = max(1, total_rows // batch_size)
    spark_df = spark_df.repartition(num_partitions)

    eval_samples = min(int(total_rows * eval_split_ratio), max_eval_samples)
    eval_df, train_df = spark_df.randomSplit([eval_samples, total_rows - eval_samples], seed=seed)

    # just in case something weird happens with Spark under the hood - ensure user can continue
    eval_df = eval_df.limit(max_eval_samples)

    print(
        f'Split {total_rows} samples into {total_rows - eval_df.count()} train samples ' \
        f'and {eval_df.count()} eval samples. Saving to {output_to}.'
    )

    train_table_path = f'{output_to}.train_split'
    eval_table_path = f'{output_to}.eval_split'

    train_df.write.format('delta').saveAsTable(train_table_path)
    eval_df.write.format('delta').saveAsTable(eval_table_path)

    print(f'Saved train data to {train_table_path} and eval data to {eval_table_path}.')

    return train_table_path, eval_table_path



"""Create a Sweep of finetuning runs"""

import logging
from typing import Dict, List, Optional, Union

from databricks.model_training.api.engine import get_return_response, run_singular_mapi_request
from databricks.model_training.api.utils import get_me
from databricks.model_training.api.validation import (SAVE_FOLDER_PATH, format_path, is_cluster_sql,
                                                      validate_create_training_run_inputs)
from databricks.model_training.types import Sweep
from databricks.model_training.types.sweep_input import SweepInput
from databricks.model_training.types.train_config import TrainTaskType

logger = logging.getLogger(__name__)

QUERY_FUNCTION = 'createFinetuneSweep'
VARIABLE_DATA_NAME = 'createFinetuneSweepData'

QUERY = f"""
mutation CreateFinetuneSweep(${VARIABLE_DATA_NAME}: CreateFinetuneSweepInput!) {{
  {QUERY_FUNCTION}({VARIABLE_DATA_NAME}: ${VARIABLE_DATA_NAME}) {{
    mlflowExperimentId
    createdAt
    sweepStatus
    reason
    finetunes {{
        id
        name
        status
        createdByEmail
        createdAt
        startedAt
        updatedAt
        completedAt
        estimatedEndTime
        reason
        details {{
            model
            taskType
            trainDataPath
            saveFolder
            evalDataPath
            trainingDuration
            learningRate
            contextLength
            dataPrepConfig
            experimentTracker
            customWeightsPath
        }}
    }}
  }}
}}"""


def create(
    model: str,
    train_data_path: str,
    register_to: str,
    *,
    experiment_name: Optional[str] = None,
    experiment_directory: Optional[str] = None,
    task_type: Optional[str] = 'INSTRUCTION_FINETUNE',
    eval_data_path: Optional[str] = None,
    custom_weights_path: Optional[str] = None,
    training_durations: Optional[Union[List[str], str]] = None,
    learning_rates: Optional[Union[List[float], float]] = None,
    context_length: Optional[int] = None,
    validate_inputs: Optional[bool] = True,
    data_prep_cluster_id: Optional[str] = None,
) -> Sweep:
    """Create a foundation model hyperparameter sweep.

    Args:
        model (str): The Hugging Face name of the base model to fine-tune.
        train_data_path (str): The full remote location of your training data.
            The format of this data depends on the task type:
            - ``INSTRUCTION_FINETUNE``: JSONL format, where each line is a
              prompt and response JSON object. Example:
              {"prompt": "What is the capital of France?", "response": "Paris"}
            - ``CHAT_COMPLETION``: JSONL format, where each line is a list of messages.
              Example:
              [
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "Hello, I need some help with my task."},
                {"role": "assistant", "content": "Yes, I can help you with that. What do you need?"}
              ]
            - ``CONTINUED_PRETRAIN``: not currently supported for Sweeps.
        register_to (str): A Unity Catalog location where the model will be registered
            after training for easy deployment. Specify a location as either
            ``<catalog_name>.<schema_name>`` or ``<catalog_name>.<schema_name>.<model_name>``.
            The former will create a model with the same name as the training run.
        experiment_name (Optional[str], optional): The name of the MLflow experiment to create. Defaults to a
            generated name if not provided.
        experiment_directory (Optional[str], optional): The path to the directory to create the MLflow experiment.
            Defaults to ``Users/<your username>``.
        task_type (Optional[str], optional): The type of task to train for. Options:
            - ``INSTRUCTION_FINETUNE``: (default): Fine-tune a model with instructions for a specific task.
            - ``CHAT_COMPLETION``: Finetune a model with chat message data.
            - ``CONTINUED_PRETRAIN``: not currently supported.
        eval_data_path (Optional[str], optional): The remote location of your evaluation data, if any.
            It must follow the same format as ``train_data_path``. Defaults to ``None``.
        custom_weights_path (Optional[str], optional): The remote location of a custom model
            checkpoint to use for the training run. If provided, these weights will be
            used instead of the model's pretrained weights. Must be an MLflow artifact path to your Composer checkpoint
            of your previously trained finetuned model on Databricks. Defaults to ``None``.
        training_durations (Optional[Union[List[str], str]], optional): A list of the total durations to launch for your
        training runs. Can be specified in:
            - Epochs (e.g. ``10ep``)
            - Tokens (e.g. ``1_000_000tok``)
            If not provided, research-backed parameters will be auto-generated based on the model you want to train.
        learning_rates (Optional[Union[List[float], float]], optional): A list of the peak learning rates to
            use for the training runs. If not provided, research-backed parameters will be auto-generated based on the
            model you want to train.
        context_length (Optional[int], optional): The maximum sequence length to use, truncating
            any data that exceeds it. Defaults to the Hugging Face model's default context
            length. Extending beyond the model's default context length is not supported.
        data_prep_cluster_id (Optional[str], optional): The cluster ID for Spark data processing.
            Required when using Delta tables as input because the underlying Delta files
            must be concatenated into a single location and converted to JSONL for
            instruction fine-tuning (IFT).

    Returns:
        Sweep: The sweep object that was created.
    """
    if experiment_directory is None:
        experiment_directory = f'/Users/{get_me()}/'
    elif experiment_directory[-1] != '/':
        experiment_directory += '/'

    train_data_path = format_path(train_data_path)
    if eval_data_path is not None:
        eval_data_path = format_path(eval_data_path)
    if custom_weights_path is not None:
        custom_weights_path = format_path(custom_weights_path)
    save_folder = SAVE_FOLDER_PATH
    if validate_inputs:
        validate_create_training_run_inputs(train_data_path, register_to, experiment_directory, experiment_name,
                                            eval_data_path, data_prep_cluster_id, custom_weights_path,
                                            TrainTaskType(task_type))

    data_prep_config: Optional[Dict[str, Union[str, bool]]] = None
    # TODO: add translations for snake to camel case
    if data_prep_cluster_id is not None:
        data_prep_config = {'clusterId': data_prep_cluster_id}
        if is_cluster_sql(data_prep_cluster_id):
            data_prep_config['useSql'] = True

    learning_rates_list: List[float] = []
    training_durations_list: List[str] = []

    if isinstance(learning_rates, float):
        learning_rates_list = [learning_rates]
    elif isinstance(learning_rates, list):
        if not all(isinstance(lr, float) for lr in learning_rates):
            raise TypeError('Found non-float in list provided for learning_rates.')
        learning_rates_list = learning_rates

    if isinstance(training_durations, str):
        training_durations_list = [training_durations]
    elif isinstance(training_durations, list):
        if not all(isinstance(td, str) for td in training_durations):
            raise TypeError('Found non-string in list provided for training_durations.')
        training_durations_list = training_durations

    sweep_input = SweepInput.from_dict({
        'models': [model],
        'task_type': task_type,
        'train_data_paths': [train_data_path],
        'model_registry_path': register_to,
        'save_folder': save_folder,
        'eval_data_path': eval_data_path,
        'training_durations': training_durations_list,
        'learning_rates': learning_rates_list,
        'context_length': context_length,
        'experiment_directory': experiment_directory,
        'experiment_name': experiment_name,
        'custom_weights_path': custom_weights_path,
        'data_prep_config': data_prep_config,
    })
    sweep_input = sweep_input.to_create_api_input()
    variables = {
        VARIABLE_DATA_NAME: sweep_input,
    }
    response = run_singular_mapi_request(
        query=QUERY,
        query_function=QUERY_FUNCTION,
        return_model_type=Sweep,
        variables=variables,
    )
    sweep = get_return_response(response)
    print(f'Created {len(sweep.finetunes)} runs.')
    return sweep



"""Provisioned throughput endpoint for model serving"""

import json
import re
import time

import requests
from mlflow.deployments import DatabricksEndpoint, get_deploy_client
from requests.exceptions import HTTPError

from databricks.model_training.api.utils import _get_host_and_token_from_env, get_me, md5_hash
from databricks.model_training.types.train_config import TrainTaskType


class ProvisionedThroughputEndpoint:
    """
    Represents a provisioned throughput endpoint of a fine-tuned model that can generate text from a
    model. The endpoint is accessed through the Databricks Model Serving REST
    API and MLFlow Deployments library.

    Args:
        uc_model_path (str): The UC location of the model to be served. Ex: 'system.ai.dbrx_instruct'
        model_version (str): The version of the model to be served. Ex: '3'
        scale_to_zero_enabled (bool): If True, the endpoint will scale to zero when not in use. Default is True.
        chunk_size_multiplier (int): The multiplier for the serving throughput chunk size. Default is 1.
        block_until_ready (bool): If True, the endpoint will block on returning until it is ready. Default is True.
    """

    def __init__(
        self,
        uc_model_path: str,
        model_version: str,
        task_type: str = 'CHAT_COMPLETION',
        scale_to_zero_enabled: bool = True,
        chunk_size_multiplier: int = 1,
        block_until_ready: bool = True,
    ) -> None:
        self._validate_inputs(uc_model_path, task_type, chunk_size_multiplier)

        model_name = uc_model_path.split('.')[-1]
        self.endpoint_name = self._generate_endpoint_name(model_name, model_version)
        self.deployment_endpoint = f'endpoints:/{self.endpoint_name}'

        self.task_type = TrainTaskType(task_type)

        # PT Config
        throughput_chunk_size = self._get_throughput_chunk_size(uc_model_path, model_version)
        self.config = {
            'served_entities': [{
                'name': self.endpoint_name,
                'entity_name': uc_model_path,
                'entity_version': model_version,
                'workload_size': 'Small',
                'workload_type': 'GPU_SMALL',
                'scale_to_zero_enabled': scale_to_zero_enabled,
                'min_provisioned_throughput': 0,
                'max_provisioned_throughput': chunk_size_multiplier * throughput_chunk_size,
            }],
        }

        self.block_until_ready = block_until_ready

        dbx_client = get_deploy_client('databricks')
        if dbx_client is None:
            raise RuntimeError('Failed to get Databricks deploy client.')
        self.client = dbx_client

        self.endpoint = self.start_pt_endpoint()

        if self.block_until_ready:
            print(f'Provisioned throughput endpoint is starting at {self.endpoint_name}. Waiting for it to be ready...')
            self.wait_for_pt_endpoint_ready()
            print(f'Provisioned throughput endpoint is ready at {self.deployment_endpoint}')
        else:
            print(f'Provisioned throughput endpoint is starting at {self.deployment_endpoint}. '
                  f"Please wait for endpoint to be ready with '{{self.__class__.wait_for_pt_endpoint_ready.__name__}}' "
                  'before use.')

    def __enter__(self) -> 'ProvisionedThroughputEndpoint':
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback) -> None:
        self.teardown_pt_endpoint()

    def __str__(self):
        return self.deployment_endpoint

    @staticmethod
    def _validate_inputs(uc_model_path: str, task_type: str, chunk_size_multiplier: int) -> None:
        """
        Validates input arguments for initialization.
        """
        model_path_regex = r'^[^/. ]+\.[^/. ]+\.[^/. ]+$'
        if not re.match(model_path_regex, uc_model_path):
            raise ValueError(
                f'Invalid uc_model_path {uc_model_path}. ' \
                "Please specify a UC location in the format '<catalog>.<schema>.<model>'. " \
                "This should be indicated under the 'Details' section on the model's page in your workspace.")

        serving_task_types = {TrainTaskType.INSTRUCTION_FINETUNE.value, TrainTaskType.CHAT_COMPLETION.value}
        if task_type not in serving_task_types:
            raise ValueError('Please specify a valid task type for the model. '
                             f'Supported task types are {serving_task_types}.')

        if chunk_size_multiplier <= 0:
            raise ValueError('Please specify a chunk size multiplier greater than 0.')

    @staticmethod
    def _generate_endpoint_name(model_name: str, model_version: str) -> str:
        """
        Helper function that generates a unique endpoint name for the PT
        endpoint. Name is unique per-user and model to avoid conflicts.
        """
        user = get_me()
        hashed_user = md5_hash(user, num_digits=8)
        # rules: must be alphanumeric w/ hyphens and less than 64 characters
        endpoint_name = '-'.join([model_name[:44], f'v{model_version[:3]}', 'eval', hashed_user])
        endpoint_name = ''.join([c if c.isalnum() or c == '-' else '-' for c in endpoint_name])
        return endpoint_name

    @staticmethod
    def _get_throughput_chunk_size(uc_model_path: str, model_version: str) -> int:
        """
        Makes an request to the `serving_endpoints` API endpoint to get the
        optimized throughput chunk size. This is required for PT serving of
        fine-tuned models.
        """
        host, token = _get_host_and_token_from_env()

        base_url = f'{host}/api/2.0/serving-endpoints'
        optimization_info_url = base_url + f'/get-model-optimization-info/{uc_model_path}/{model_version}'
        request_headers = {
            'Context-Type': 'text/json',
            'Authorization': f'Bearer {token}',
        }
        response = requests.get(url=optimization_info_url, headers=request_headers, timeout=10)
        if response.status_code != 200:
            raise HTTPError(f'Failed to retrieve throughput chunk size. Error {response.status_code}: {response.text}')

        throughput_information = response.json()
        throughput_chunk_size = throughput_information.get('throughput_chunk_size')

        # If `optimizable` is False, then we can't get `throughput_chunk_size`
        # because the model isn't supported by Databricks Model Serving.
        if throughput_chunk_size is None:
            optimizable = throughput_information.get('optimizable', False)
            if not optimizable:
                raise ValueError('Please use a model that is currently supported by Databricks Model serving: '
                                 'https://docs.databricks.com/en/machine-learning/foundation-models/index.html'
                                 '#provisioned-throughput-foundation-model-apis')
            raise ValueError('Could not retrieve throughput chunk size from '
                             '`serving-endpoints/model-optimization-info`. '
                             f'{json.dumps(throughput_information)}')

        return throughput_chunk_size

    def is_endpoint_ready(self) -> bool:
        """
        Returns True if the endpoint is scaled from zero and ready to serve
        requests. Returns False otherwise.
        """
        return self.endpoint.state.ready == 'READY'

    def start_pt_endpoint(self) -> DatabricksEndpoint:
        """
        Spins up a provisioned throughput endpoint on behalf of the user's
        provided model.
        """
        try:
            existing_endpoint = self.client.get_endpoint(self.endpoint_name)
            print(f'Endpoint {self.endpoint_name} already exists. Skipping endpoint creation. '
                  f"Please wait for endpoint to be ready with '{self.__class__.wait_for_pt_endpoint_ready.__name__}' "
                  'before use.')
            return existing_endpoint
        except HTTPError as e:
            if e.response.status_code != 404:
                raise e

        # Create the endpoint through a POST request
        endpoint = self.client.create_endpoint(name=self.endpoint_name, config=self.config)

        return endpoint

    def wait_for_pt_endpoint_ready(self, timeout_mins: int = 30, check_interval_secs: int = 30):
        """
        Waits for the PT endpoint to be ready before returning to the user.
        """
        t0 = time.time()
        timeout_secs = timeout_mins * 60
        while time.time() - t0 < timeout_secs:
            state = self.endpoint.state
            if state['ready'] == 'READY':
                return
            if state['config_update'] == 'UPDATE_FAILED':
                raise RuntimeError('Endpoint update failed. Please validate your served model endpoint and spin'
                                   ' it down manually if necessary.')
            if state['config_update'] == 'UPDATE_CANCELED':
                raise RuntimeError('Endpoint update was canceled. Please validate your served model endpoint and spin'
                                   ' it down manually if necessary.')
            time.sleep(check_interval_secs)

        raise TimeoutError(f'Provisioned throughput endpoint was not ready within {timeout_mins} minutes.')

    def teardown_pt_endpoint(self):
        """
        Tears down the provisioned throughput endpoint once the user's evaluation is complete.
        """
        if self.is_endpoint_ready():
            self.client.delete_endpoint(self.endpoint_name)
            print(f'Endpoint {self.endpoint_name} has been torn down.')
        else:
            print(f'Endpoint {self.endpoint_name} is not ready. Skipping teardown.')

    def query(self, input_data: dict) -> str:
        """Queries the PT endpoint with the given input"""
        if not self.is_endpoint_ready():
            raise RuntimeError('Provisioned throughput endpoint is not ready.'
                               'Please start it with `start_pt_endpoint`.')

        chat_response = self.client.predict(
            endpoint=self.endpoint_name,
            inputs=input_data,
        )
        if not chat_response:
            raise RuntimeError(f'Failed to generate completion for input: {input_data}')

        return chat_response['choices'][0]['message']['content']

    def generate_completion(self, prompt: str, temperature: float = 1.0, max_tokens: int = 128) -> str:
        """Uses PT endpoint to generate completion for a given prompt string"""
        if self.task_type == TrainTaskType.INSTRUCTION_FINETUNE:  # completions endpoint
            inputs = {
                'prompt': prompt,
                'temperature': temperature,
                'max_tokens': max_tokens,
            }
        elif self.task_type == TrainTaskType.CHAT_COMPLETION:  # chat endpoint
            inputs = {
                'messages': [{
                    # TODO: add a system prompt here?
                    'role': 'user',
                    'content': prompt,
                }],
                'temperature': temperature,
                'max_tokens': max_tokens,
            }
        else:
            raise ValueError(f'Unsupported task type configured: {self.task_type.value}')

        return self.query(inputs)



"""Sweep Status"""
from enum import Enum
from typing import Union

from mcli.utils.utils_string_functions import camel_case_to_snake_case, snake_case_to_camel_case

__all__ = ['SweepStatus']


class SweepStatus(Enum):
    """Possible statuses of a sweep
    """
    # If all runs in sweep are pending/queued
    PENDING = 'PENDING'

    # If at least one run in sweep has started training
    RUNNING = 'RUNNING'

    # User manually canceled a sweep or all runs in a sweep
    CANCELED = 'CANCELED'

    # There is 1+ failed run in sweep
    FAILED = 'FAILED'

    # If all runs in sweep are completed
    COMPLETED = 'COMPLETED'

    def __str__(self) -> str:
        return self.value

    @classmethod
    def from_string(cls, sweep_status: Union[str, 'SweepStatus']) -> 'SweepStatus':
        """Convert a string to a valid SweepStatus Enum

        If the sweep status string is not recognized, will return SweepStatus.FAILED
        instead of raising a KeyError
        """
        if isinstance(sweep_status, SweepStatus):
            return sweep_status

        default = SweepStatus.FAILED  # Defaulting to FAILED for unrecognized statuses
        try:
            key = camel_case_to_snake_case(sweep_status).upper()
            return cls[key]
        except TypeError:
            return default
        except KeyError:
            return default

    @property
    def display_name(self) -> str:
        return snake_case_to_camel_case(self.value, capitalize_first=True)



"""
Utilities for formatting foundation model training objects.
"""
from typing import Any, Dict, List
from urllib.parse import urljoin

from databricks.model_training.api.utils import _get_host_and_token_from_env, is_running_in_databricks_notebook


def get_mlflow_experiment_link_html(experiment_id: str, link_text: str = 'experiment') -> str:
    """
    Helper function to get a link for the MLflow experiment using the given experiment ID.
    """
    if is_running_in_databricks_notebook():
        host = ''
    else:
        host, _ = _get_host_and_token_from_env()
    path = urljoin(host, f'/ml/experiments/{experiment_id}')
    return f"<a href={path} class target='_blank'>{link_text}</a>"


def get_mlflow_run_link_html(experiment_id: str, run_id: str, link_text: str = 'run') -> str:
    """
    Helper function to get a link for the MLflow run using the given experiment ID and run ID.
    """
    if is_running_in_databricks_notebook():
        host = ''
    else:
        host, _ = _get_host_and_token_from_env()
    path = urljoin(host, f'/ml/experiments/{experiment_id}/runs/{run_id}')
    return f"<a href={path} class target='_blank'>{link_text}</a>"


def create_horizontal_html_table(data: List[Dict[str, Any]], key_to_label: Dict[str, str]) -> str:
    """
    Helper function to generate a horizontal HTML table.

    Parameters
    ----------
    data : List[Dict[str, Any]]
        A list of dictionaries where each dictionary represents a row.
    key_to_label : Dict[str, str]
        A dictionary mapping the keys in the data to the desired column headers.

    Returns
    -------
    str
        An HTML table as a string with headers in the first row and data in the subsequent rows.
    """
    res = []
    res.append("<table border=\"1\" class=\"dataframe\">")

    # Header row (labels from key_to_label)
    res.append('<thead>')
    res.append('<tr>')
    for key in key_to_label:
        res.append(f'<th style="text-align: left;">{key_to_label[key]}</th>')
    res.append('</tr>')
    res.append('</thead>')

    # Body rows (data)
    res.append('<tbody>')
    for row in data:
        res.append('<tr>')
        for key in key_to_label:
            value = row.get(key, None)
            res.append(f'<td style="text-align: left;">{value if value is not None else "-"}</td>')
        res.append('</tr>')
    res.append('</tbody>')

    res.append('</table>')
    return '\n'.join(res)


def create_vertical_html_table(data: List[Dict[str, Any]], key_to_label: Dict[str, str]) -> str:
    """
    Helper function to generate a vertical HTML table.

    Parameters
    ----------
    data : List[Dict[str, Any]]
        A list of dictionaries where each dictionary represents a row.
    key_to_label : Dict[str, str]
        A dictionary mapping the keys in the data to the desired row headers.

    Returns
    -------
    str
        An HTML table as a string with headers in the first column and values in the remaining columns.
    """
    res = []
    res.append("<table border='1' class='dataframe'>")
    res.append('<tbody>')

    # Each key becomes a row, with its values spread out across columns
    for key, label in key_to_label.items():
        res.append('<tr>')
        res.append(f'<th style="text-align: left;">{label}</th>')
        for row in data:
            value = row.get(key, None)
            res.append(f'<td style="text-align: left;">{value if value is not None else "-"}</td>')
        res.append('</tr>')

    res.append('</tbody>\n</table>')
    return '\n'.join(res)



"""
Get model training sweep
"""
from datetime import datetime
from typing import List, Optional, Union

from databricks.model_training.api.engine import get_return_response, run_plural_mapi_request
from databricks.model_training.api.validation import parse_datetime
from databricks.model_training.types.common import ObjectList
from databricks.model_training.types.sweep import Sweep
from databricks.model_training.types.sweep_status import SweepStatus

DEFAULT_SWEEP_LIMIT = 5

QUERY_FUNCTION = 'getFinetuneSweeps'
VARIABLE_DATA_NAME = 'getFinetuneSweepsData'

QUERY = f"""
query GetFinetuneSweeps(${VARIABLE_DATA_NAME}: GetFinetuneSweepsInput!) {{
  {QUERY_FUNCTION}({VARIABLE_DATA_NAME}: ${VARIABLE_DATA_NAME}) {{
    mlflowExperimentId
    createdAt
    sweepStatus
    reason
    finetunes {{
        id
        name
        status
        createdByEmail
        createdAt
        updatedAt
        startedAt
        completedAt
        reason
        estimatedEndTime
        details {{
            model
            taskType
            trainDataPath
            saveFolder
            evalDataPath
            trainingDuration
            learningRate
            contextLength
            dataPrepConfig
            experimentTracker
            customWeightsPath
        }}
     }}
  }}
}}"""


def get(
    experiment_ids: Optional[Union[str, List[str]]] = None,
    *,
    statuses: Optional[Union[List[str], List[SweepStatus]]] = None,
    before: Optional[Union[str, datetime]] = None,
    after: Optional[Union[str, datetime]] = None,
    limit: Optional[int] = DEFAULT_SWEEP_LIMIT,
) -> ObjectList[Sweep]:
    """
    Get sweep details filtered by MLflow experiment IDs, statuses, and creation date.

    Args:
        experiment_ids (Optional[Union[str,List[str]]]): List of MLflow experiment IDs or a single experiment
            ID to filter the sweeps by.
        statuses (Optional[Union[List[str], List[RunStatus]]]): Filter by the status of the sweeps.
        before (Optional[Union[str, datetime]]): Filter sweeps created before a specific date in user's local time.
        after (Optional[Union[str, datetime]]): Filter sweeps created after a specific date in user's local time.

    Returns:
        ObjectList[Sweep]: The sweep details associated with the given filters.
    """
    filters = {}
    if experiment_ids:
        if isinstance(experiment_ids, str):
            experiment_ids = [experiment_ids]
        filters = {'mlflowExperimentId': {'in': experiment_ids}}
    if statuses:
        filters['sweepStatus'] = {'in': [s.value if isinstance(s, SweepStatus) else s for s in statuses]}
    if before or after:
        date_filters = {}
        if before:
            date_filters['lt'] = before.astimezone().isoformat() if isinstance(before,
                                                                               datetime) else parse_datetime(before)
        if after:
            date_filters['gte'] = after.astimezone().isoformat() if isinstance(after,
                                                                               datetime) else parse_datetime(after)
        filters['createdAt'] = date_filters

    variables = {
        VARIABLE_DATA_NAME: {
            'filters': filters,
            'limit': limit
        },
    }

    response = run_plural_mapi_request(
        query=QUERY,
        query_function=QUERY_FUNCTION,
        return_model_type=Sweep,
        variables=variables,
    )

    return get_return_response(response)



"""
For training foundation models
"""
from . import evaluation
from .cancel_runs import cancel, cancel_run, cancel_runs
from .create import create
from .get import get
from .get_checkpoints import get_checkpoints
from .get_models import get_models
from .get_run_events import get_run_events
from .get_runs import get_run, get_runs

__all__ = [
    'get_checkpoints', 'get_models', 'cancel', 'cancel_run', 'cancel_runs', 'create', 'get', 'get_run', 'get_runs',
    'get_run_events', 'evaluation'
]



"""Utilities for logging evaluation metrics"""

from typing import Optional, TypedDict


class EvalMetricDict(TypedDict):
    metric_name: str
    score: float
    greater_is_better: bool
    baseline_score: Optional[float]



"""
A Databricks sweep.
"""
import logging
from dataclasses import dataclass
from datetime import datetime
from http import HTTPStatus
from typing import Any, Dict, List, Optional, Tuple

from mcli import MAPIException
from mcli.api.schema.generic_model import DeserializableModel, convert_datetime
from mcli.utils.utils_string_functions import camel_case_to_snake_case

from databricks.model_training.types.sweep_status import SweepStatus

from .training_run import TrainingRun
from .utils import create_horizontal_html_table, get_mlflow_experiment_link_html, get_mlflow_run_link_html

logger = logging.getLogger(__name__)


@dataclass()
class Sweep(DeserializableModel):
    """A Databricks sweep object.

        Args:
            mlflow_experiment_id (str): The MLflow experiment ID.
            created_at (datetime): The timestamp when the sweep was created.
            sweep_status (SweepStatus): The status of the sweep.
            finetunes (List[Finetune]): The list of Finetune runs.
            reason (Optional[str], optional): The optional reason to expand on the sweep status. Defaults to None.
    """
    mlflow_experiment_id: str
    created_at: datetime
    sweep_status: SweepStatus
    finetunes: List[TrainingRun]
    reason: Optional[str] = None

    _required_properties: Tuple[str, ...] = tuple([
        'mlflowExperimentId',
        'createdAt',
        'sweepStatus',
        'finetunes',
    ])

    @classmethod
    def from_mapi_response(cls, response: Dict[str, Any]) -> 'Sweep':
        """Load the sweep from MAPI response."""
        missing = set(cls._required_properties) - set(response.keys())
        if missing:
            raise MAPIException(status=HTTPStatus.BAD_REQUEST,
                                message='Missing required key(s) in response to deserialize Sweep '
                                f'object: {", ".join(missing)}')
        args = {camel_case_to_snake_case(key): value for key, value in response.items()}
        return cls(
            mlflow_experiment_id=args['mlflow_experiment_id'],
            created_at=convert_datetime(args['created_at']),
            sweep_status=SweepStatus.from_string(args['sweep_status']),
            reason=args['reason'] if 'reason' in args else None,
            finetunes=[TrainingRun.from_mapi_response(finetune) for finetune in args['finetunes']],
        )

    def __iter__(self):
        """Iterate over finetunes."""
        return iter(self.finetunes)

    @property
    def _sweep_status_str(self) -> str:
        """Get the sweep status as a string.
        """
        if self.reason:
            return f'{self.sweep_status} ({self.reason})'
        return str(self.sweep_status)

    @property
    def _mlflow_experiment_link_html(self) -> str:
        """
        Get a hyperlinked experiment ID.
        This link will open the MLflow experiment page in a new tab.

        Raises:
            ValueError: If the experiment ID is not set
        """
        if self.mlflow_experiment_id is None:
            raise ValueError('Experiment ID is not set')
        return get_mlflow_experiment_link_html(self.mlflow_experiment_id, self.mlflow_experiment_id)

    @property
    def _created_at_str(self) -> str:
        """Get the created at as a string.
        """
        return self.created_at.strftime('%Y-%m-%d %H:%M:%S')

    @property
    def _sweep_finetune_table(self):
        """Display table with sweep finetunes."""
        pruned_finetunes = []
        for finetune in self.finetunes:
            pruned_finetunes.append({
                'name': get_mlflow_run_link_html(self.mlflow_experiment_id, finetune.run_id, finetune.name),
                'status': finetune.status,
                'training_duration': finetune.training_duration,
                'learning_rate': finetune.learning_rate,
            })
        sorted_finetunes = sorted(pruned_finetunes, key=lambda x: (x['training_duration'], x['learning_rate']))
        key_to_label = {key: ' '.join([word.title() for word in key.split('_')]) for key in pruned_finetunes[0].keys()}
        return create_horizontal_html_table(sorted_finetunes, key_to_label)

    def _repr_html_(self) -> str:
        """Display the sweep as HTML.
            """
        data = {
            'mlflow_experiment_id': self._mlflow_experiment_link_html,
            'created_at': self._created_at_str,
            'status': self._sweep_status_str,
            'runs': self._sweep_finetune_table
        }
        key_to_label = {key: key.replace('_', ' ').title() for key in data}
        return create_horizontal_html_table([data], key_to_label)



""" Serving utilities """
from typing import Dict, Optional, Tuple

"""
Mapping from supported FT model name to the corresponding model path and version.
Models that are not supported by serving should be set to None.
Some base models are not enabled for serving. Those models are substituted with their instruct versions.
"""
FT_MODEL_PATH_MAPPING: Dict[str, Optional[Tuple[str, str]]] = {
    'databricks/dbrx-base': ('system.ai.dbrx_base', '2'),
    'databricks/dbrx-instruct': ('system.ai.dbrx_instruct', '3'),
    'meta-llama/Llama-3.2-1B': ('system.ai.llama_v3_2_1b_instruct', '2'),  # SUBSTITUTION
    'meta-llama/Llama-3.2-1B-Instruct': ('system.ai.llama_v3_2_1b_instruct', '4'),
    'meta-llama/Llama-3.2-3B': ('system.ai.llama_v3_2_3b_instruct', '2'),  # SUBSTITUTION
    'meta-llama/Llama-3.2-3B-Instruct': ('system.ai.llama_v3_2_3b_instruct', '4'),
    'meta-llama/Meta-Llama-3.1-405B': ('system.ai.meta_llama_v3_1_405b', '1'),
    'meta-llama/Meta-Llama-3.1-405B-Instruct': ('system.ai.meta_llama_v3_1_405b_instruct_fp8', '1'),
    'meta-llama/Meta-Llama-3.1-70B': ('system.ai.meta_llama_v3_1_70b_instruct', '2'),  # SUBSTITUTION
    'meta-llama/Meta-Llama-3.1-70B-Instruct': ('system.ai.meta_llama_v3_1_70b_instruct', '2'),
    'meta-llama/Meta-Llama-3.1-8B': ('system.ai.meta_llama_v3_1_8b_instruct', '2'),  # SUBSTITUTION
    'meta-llama/Meta-Llama-3.1-8B-Instruct': ('system.ai.meta_llama_v3_1_8b_instruct', '4'),
    'meta-llama/Meta-Llama-3-70B': ('system.ai.meta_llama_3_70b', '1'),
    'meta-llama/Meta-Llama-3-70B-Instruct': ('system.ai.meta_llama_3_70b_instruct', '2'),
    'meta-llama/Meta-Llama-3-8B': ('system.ai.meta_llama_3_8b', '1'),
    'meta-llama/Meta-Llama-3-8B-Instruct': ('system.ai.meta_llama_3_8b_instruct', '1'),
    'meta-llama/Llama-2-7b-hf': None,
    'meta-llama/Llama-2-13b-hf': None,
    'meta-llama/Llama-2-70b-hf': None,
    'meta-llama/Llama-2-7b-chat-hf': None,
    'meta-llama/Llama-2-13b-chat-hf': None,
    'meta-llama/Llama-2-70b-chat-hf': None,
    'codellama/CodeLlama-7b-hf': None,
    'codellama/CodeLlama-13b-hf': None,
    'codellama/CodeLlama-34b-hf': None,
    'codellama/CodeLlama-7b-Instruct-hf': None,
    'codellama/CodeLlama-13b-Instruct-hf': None,
    'codellama/CodeLlama-34b-Instruct-hf': None,
    'codellama/CodeLlama-7b-Python-hf': None,
    'codellama/CodeLlama-13b-Python-hf': None,
    'codellama/CodeLlama-34b-Python-hf': None,
    'mistralai/Mistral-7B-v0.1': ('system.ai.mistral_7b_v0_1', '2'),
    'mistralai/Mistral-7B-Instruct-v0.2': ('system.ai.mistral_7b_instruct_v0_2', '2'),
    'mistralai/Mixtral-8x7B-v0.1': ('system.ai.mixtral_8x7b_v0_1', '2'),
}


def get_base_model_details(model_name: str) -> Tuple[str, str]:
    """
    Get the model path for the given model name

    Args:
        model_name (str): The FT base model name

    Returns:
        Tuple[str, str]: The model path and version
    """
    model_details = FT_MODEL_PATH_MAPPING.get(model_name)
    if model_details is None:
        raise ValueError(f"Model '{model_name}' is not supported.")
    return model_details



"""Utility functions for the API."""

import hashlib
import logging
import os
from typing import Tuple

from databricks.sdk import WorkspaceClient
from IPython import get_ipython  # type: ignore

logger = logging.getLogger(__name__)

_LOCAL_DEV_CONFIG_PROFILE = 'DBX_DEV_LOCAL'
_DATABRICKS_CONFIG_PROFILE_ENV = 'DATABRICKS_CONFIG_PROFILE'
_TEST_CONFIG_PROFILE = 'DBX_DEV_TEST'


def is_running_in_databricks_notebook() -> bool:
    try:
        shell = get_ipython().__class__.__name__  # pylint: disable = undefined-variable
        if shell == 'DatabricksShell':
            return True
        return False
    except Exception:  # pylint: disable = broad-except
        return False


def get_me() -> str:
    """
    Get who is currently logged in.

    Returns:
        str: The name of the current user.
    """
    # TODO remove, only used for testing
    if os.environ.get(_DATABRICKS_CONFIG_PROFILE_ENV, '') in [
            _LOCAL_DEV_CONFIG_PROFILE,
            _TEST_CONFIG_PROFILE,
    ]:
        return 'me'

    w = WorkspaceClient()
    me = w.current_user.me().user_name or ''
    if not me:
        raise EnvironmentError('Could not determine the current user. Please check your environment.')
    logger.debug(f'You are {me}')
    return me


def _get_host_and_token_from_env() -> Tuple[str, str]:
    """
    Get the host and token from the environment

    In a databricks notebook, the host will be the underlying region and env (i.e. oregon.staging.databricks.com)
    """
    if os.environ.get(_DATABRICKS_CONFIG_PROFILE_ENV, '') == _TEST_CONFIG_PROFILE:
        return 'test', 'test'
    w = WorkspaceClient()
    if is_running_in_databricks_notebook():
        ctx = w.dbutils.entry_point.getDbutils().notebook().getContext()
        host = ctx.apiUrl().get()
        token = ctx.apiToken().get()
    else:
        host = w.config.host
        token = w.config.token
    if not host or not token:
        raise EnvironmentError('Could not find root URL and/or token. Please check your environment.')
    return host, token


def get_cluster_id() -> str:
    """
    Get the cluster ID from the environment
    """
    if os.environ.get(_DATABRICKS_CONFIG_PROFILE_ENV, '') == _TEST_CONFIG_PROFILE:
        return 'test'
    w = WorkspaceClient()
    if is_running_in_databricks_notebook():
        ctx = w.dbutils.notebook.entry_point.getDbutils().notebook().getContext()
        cluster_id = ctx.clusterId().get()
    else:
        cluster_id = w.config.cluster_id
    return cluster_id


def md5_hash(s: str, num_digits=8) -> str:
    return hashlib.md5(s.encode()).hexdigest()[:num_digits]
